from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypedDict

import numpy as np
import xarray as xr
from xarray import Dataset, DataTree, open_datatree

from eopf import EOLogging

if TYPE_CHECKING:
    from xarray.core.datatree import DatasetView

# Keyword arguments for xarray.open_dataset
OpenDatasetKwargs = TypedDict(
    "OpenDatasetKwargs",
    {
        "chunks": dict[str, int],
        "engine": str,
        "drop_variables": list[str],
    },
)
# Keyword arguments for merging xarray datasets
MergeKwargs = TypedDict(
    "MergeKwargs",
    {
        "compat": Literal["identical"],
        "fill_value": dict[str, Any] | None,
        "combine_attrs": Literal["drop_conflicts"],
    },
)

GROUPS_TO_SKIP = [
    "/conditions/meteo",  # L1
    "/conditions/meteorology/cams",  # L2
    "/conditions/meteorology/ecmwf",  # L2
]
OPEN_DATASET_KWARGS: OpenDatasetKwargs = {
    "chunks": {},
    "engine": "zarr",
    "drop_variables": [
        "mean_vaa",  # L1
        "mean_vza",  # L1
        "mean_sun_angles",  # L2
        "mean_viewing_incidence_angles",  # L2
    ],  # TODO: drop o re-compute after combining all tiles?
}
UINT_NAN = 0  # Placeholder fill value for unsigned integer types
GEOMETRY_GROUP = "/conditions/geometry"
# Attributes to ignore when comparing metadata
INCONSISTENT_ATTRS = [
    "other_metadata",
    "stac_discovery",
    "proj:bbox",
    "proj:transform",
]


def _get_spatial_coords(ds: Dataset) -> list[str]:
    """
    Extracts spatial coordinates from a dataset.

    Parameters
    ----------
    ds : Dataset
        Input dataset.

    Returns
    -------
    list[str]
        List of spatial coordinate names (e.g., ["x10", "y10"]).
    """
    return [coord for coord in ds.coords if isinstance(coord, str) and coord.startswith(("x", "y"))]


def _find_slicers(paths: list[str]) -> dict[str, dict[str, slice]]:
    """
    Determines spatial slices for each tile based on its coordinates.

    Parameters
    ----------
    paths : list[str]
        List of Zarr tile paths.

    Returns
    -------
    dict[str, dict[str, slice]]
        Mapping from tile path to spatial slicing bounds.
    """
    tiles = []
    for path in paths:
        # Open the geometry group of the dataset to get spatial bounds
        ds = xr.open_dataset(path, group=GEOMETRY_GROUP, **OPEN_DATASET_KWARGS)
        coords = _get_spatial_coords(ds)
        # Select endpoints and sort by coordinates for bounds calculation
        ds = ds[coords].isel({coord: [0, -1] for coord in coords}).sortby(coords)
        variables = {}
        for coord in coords:
            variables[coord] = ds[coord].mean(coord, keepdims=True)
            variables[f"{coord}_bounds"] = xr.DataArray(ds[coord].values, dims="bounds")
        ds = xr.Dataset(variables)
        tiles.append(ds.expand_dims(tile=[path]))
    # Stack all tiles into one dataset for global comparison
    ds = xr.concat(tiles, "tile")
    slicers = {}
    for ds_tile in tiles:
        path = ds_tile["tile"].item()
        ds_tile = ds_tile.squeeze("tile")
        slicer = {}
        for coord in _get_spatial_coords(ds_tile):
            size = ds_tile[f"{coord}_bounds"].diff("bounds")
            bounds = list(ds_tile[f"{coord}_bounds"].values)
            for label in ("upper", "lower"):
                # Ignored type as it doesn't detect that we have upper and lower literal
                diff = ds[coord].diff(coord, label=label)  # type: ignore
                try:
                    distance = diff.sel({coord: ds_tile[coord]})
                except KeyError:
                    continue
                else:
                    if distance >= size:
                        continue
                    shift = (size - distance).item() / 2
                    if label == "lower":
                        bounds[1] -= shift + 0.5  # exclusive
                    else:
                        bounds[0] += shift  # inclusive
            slicer[coord[0]] = slice(*bounds)
        slicers[path] = slicer

    return slicers


def _preprocess(ds: Dataset, slicers: dict[str, dict[str, slice]]) -> Dataset:
    """
    Applies spatial slicing and sorting to a dataset.

    Parameters
    ----------
    ds : Dataset
        The dataset to preprocess.
    slicers : dict[str, dict[str, slice]]
        Slicers per tile path and coordinate.

    Returns
    -------
    Dataset
        Preprocessed dataset.
    """
    path = ds.encoding["source"]
    indexers = {dim: slicer for dim in ds.dims if isinstance(dim, str) and (slicer := slicers[path].get(dim[0]))}
    return ds.sortby(list(indexers)).sel(indexers)


# ----------------------------------------
# Main Tile-Combining Function
# ----------------------------------------


def open_and_combine_tiles(input_dir: str, allow_missing_tiles: bool = False) -> DataTree:
    """
    Opens and combines multiple spatial tiles into a single hierarchical DataTree.

    Parameters
    ----------
    input_dir : str
        Path to a directory containing Zarr tiles.
    allow_missing_tiles : bool
        Whether to allow missing data and fill with defaults.

    Returns
    -------
    DataTree
        Combined data structure of all tiles.
    """
    logger = EOLogging().get_logger("eopf.computing.merge")
    logger.info(f"Searching for products in {input_dir}")
    logger.info(f"Allow missing tiles: {allow_missing_tiles}")
    paths = list(map(str, Path(input_dir).iterdir()))
    logger.debug(f"{len(paths)} products found")
    slicers = _find_slicers(paths)

    sample_dict = open_datatree(paths[0], **OPEN_DATASET_KWARGS).to_dict()
    output_dict = {}
    for group, ds in sample_dict.items():
        if group in GROUPS_TO_SKIP:
            continue

        # Preprocess all tiles for this group to apply slicing
        tiles = [
            _preprocess(
                xr.open_dataset(path, group=group, **OPEN_DATASET_KWARGS),
                slicers,
            )
            for path in paths
        ]
        # Define merge options
        merge_kwargs: MergeKwargs = {
            "compat": "identical",
            "fill_value": (
                {
                    var: np.array(UINT_NAN if da.dtype.kind == "u" else np.nan, da.dtype)
                    for var, da in ds.data_vars.items()
                }
                if allow_missing_tiles
                else None
            ),
            "combine_attrs": "drop_conflicts",
        }

        # Process each resolution separately
        # Merge scalar variables (non-spatial)
        spatial_coords = _get_spatial_coords(ds)
        datasets = [xr.merge([tile.drop_dims(spatial_coords, errors="ignore") for tile in tiles], **merge_kwargs)]
        # Merge spatial variables, grouped by resolution
        for suffix in {coord[1:] for coord in spatial_coords}:
            dims_to_drop = [coord for coord in spatial_coords if not coord[1:] == suffix]
            ds_combined = xr.combine_by_coords(
                [tile.drop_dims(dims_to_drop, errors="ignore") for tile in tiles],
                **merge_kwargs,
                data_vars="minimal",
                coords="minimal",
            )
            if not isinstance(ds_combined, Dataset):
                raise Exception("")
            datasets.append(ds_combined)
        ds = xr.merge(datasets, **merge_kwargs)

        # Apply chunking for efficient storage
        chunks = {}
        for da in ds.data_vars.values():
            chunks.update(da.encoding["preferred_chunks"])
        output_dict[group] = ds.chunk(chunks)

    return DataTree.from_dict(output_dict)


# ----------------------------------------
# Sanity Check
# ----------------------------------------


def _sanity_check(path: Path, dt: DataTree) -> None:
    """
    Internal: checks if a single tile matches its representation in the DataTree.

    Parameters
    ----------
    path : Path
        Path to the original tile.
    dt : DataTree
        Combined DataTree.
    """
    dt_tile = open_datatree(path, engine="zarr")
    for group, ds_tile in dt_tile.to_dict().items():
        if group in GROUPS_TO_SKIP:
            if group in dt.groups:
                raise Exception("")
            continue

        # Variables
        ds: DatasetView = dt[group].dataset
        if not set(ds.dims) == set(ds_tile.dims):
            raise Exception("")

        if not set(ds.coords) == set(ds_tile.coords):
            raise Exception("")

        if not set(ds.data_vars) == set(ds_tile.data_vars) - set(OPEN_DATASET_KWARGS["drop_variables"]):
            raise Exception("")

        # Resolution
        if group != GEOMETRY_GROUP:
            for dim in ds.dims:
                if not isinstance(dim, str) or dim[0] not in ("x", "y"):
                    continue
                (delta,) = set(ds[dim].diff(dim).values)
                digits = "".join([s for s in dim if s.isdigit()] or [s for s in group.rsplit("/")[-1] if s.isdigit()])
                if not delta == int(digits):
                    raise Exception("")

        # Attributes
        if not ds.attrs == {k: v for k, v in ds_tile.attrs.items() if k not in INCONSISTENT_ATTRS}:
            raise Exception("")
        for var, da in ds.variables.items():
            da_tile = ds_tile[var]
            if not da.dtype == da_tile.dtype:
                raise Exception("")
            if not da.attrs == {k: v for k, v in da_tile.attrs.items() if k not in INCONSISTENT_ATTRS}:
                raise Exception("")


def sanity_check(input_dir: str, dt: DataTree | None = None) -> None:
    """
    Runs sanity checks to validate that combined DataTree matches individual tiles.

    Parameters
    ----------
    input_dir : str
        Path to directory containing tiles.
    dt : DataTree | None
        Optional precomputed DataTree. If None, it will be recomputed.
    """
    if dt is None:
        dt = open_and_combine_tiles(input_dir)
    for path in Path(input_dir).iterdir():
        _sanity_check(path, dt)
